#brailleDisplayDrivers/papenmeier.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Author: Tobias Platen (nvda@lists.thm.de)
#minor changes by Halim Sahin (nvda@lists.thm.de)
import time
import wx
import braille
from logHandler import log


import inputCore
import ftdi2
import struct

#for bluetooth
import hwPortUtils
import serial

#for brxcom
import ctypes as c
import _winreg
import winUser


#timer intervalls used by the driver
KEY_CHECK_INTERVAL = 50
BLUETOOTH_INTERVAL = 5000

#Control Flow
STX = 0x02 #Start of Text
ETX = 0x03 #End of Text

#Control Messages
AUTOID = 0x42
BRAILLE = 0x43

#Timeout for bluetooth
BLUETOOTH_TIMEOUT = 0.2


def brl_auto_id(): 
	"""send auto id command to braille display"""
	return struct.pack('bbbbb',STX,AUTOID,0x50,0x50,ETX)
        #device will respond with a message that allows identification of the display


def brl_out(data,nrk,nlk,nv):
	"""write data to braille cell with nv vertical cells, nrk cells right and nlk cells left
	   some papenmeier displays have vertical cells, other displays have dummy cells with keys
	""" 
	ret = struct.pack('BB', STX, BRAILLE) #STX,COMMAND BRAILLE
	d2 = len(data)
	#reserve space for braille displays with dummy cells
	d2+=nv
	d2+=2*nlk
	d2+=2*nrk
	a = 0x50|(d2 & 0x0F)
	b = 0x50|(d2 >> 4)
	#write length to stream 
	ret += struct.pack('BB',b,a)
	#fill dummy bytes (left,vertical)
	for i in range(0,nv):
		ret += struct.pack('BB',0x30,0x30)
	for i in range(0,nlk):
		ret += struct.pack('BBBB',0x30,0x30,0x30,0x30)
	#swap dot bits        
	for d in data:
		d2 = 0
		if(d & 1): d2|=128
		if(d & 2): d2|=64
		if(d & 4): d2|=32
		if(d & 8): d2|=16
		if(d & 16): d2|=8
		if(d & 32): d2|=4
		if(d & 64): d2|=2
		if(d & 128): d2|=1
		a = 0x30|(d2 & 0x0F)
		b = 0x30|(d2 >> 4)
		ret += struct.pack('BB',b,a)
	#fill dummy bytes on (right)        
	for i in range(0,nrk):
		ret += struct.pack('BBBB',0x30,0x30,0x30,0x30)
	#ETX        
	ret += struct.pack('B',ETX)
	return ret

def brl_poll(dev): 
	"""read sequence from braille display,non blocking"""
	q = dev.inWaiting()
	status = ''
	if(q>0):
		d=dev.read(1)
		if(ord(d[0])==STX):#first char must be an STX
			while(dev.inWaiting()==0): pass
			d=dev.read(1)
			while(ord(d[0])!=ETX): #wait until ETX
				status += d[0]
				while(dev.inWaiting()==0): pass
				d=dev.read(1)
	return status
       
def brl_decode_trio(keys): 
	"""decode routing keys on Trio"""
	if(keys[0:3]=='KP_' ): #KEYSTATE CHANGED EVENT on Trio, not Braille keys
		keys = keys[3:]
		i = 0
		j = []
		for k in keys:
			a= ord(k)&0x0F
			#convert bitstream to list of indexes
			if(a & 1): j+= [i+3]
			if(a & 2): j+= [i+2]
			if(a & 4): j+= [i+1]
			if(a & 8): j+= [i]
			i +=4
		return j
	return -1


def brl_decode_keys_A(data,start,voffset):
	"""decode routing keys non Trio devices"""
	n = start                           #key index iterator 
	j=  []
	shift = 0
	for i in range(0,len(data)):	#byte index
		if(i%2==0):
			a= ord(data[i])&0x0F		#n+4,n+3
			b= ord(data[i+1])&0x0F	#n+2,n+1
			#convert bitstream to list of indexes
			if(n > 26): shift=voffset
			if(b & 1): j += [n+0-shift]
			if(b & 2): j += [n+1-shift]
			if(b & 4): j += [n+2-shift]
			if(b & 8): j += [n+3-shift]
			if(a & 1): j += [n+4-shift]
			if(a & 2): j += [n+5-shift]
			if(a & 4): j += [n+6-shift]
			if(a & 8): j += [n+7-shift]
			n+=8	
	return j

def brl_decode_key_names_A_repeat(keys,driver):
	"""translate key names for protocol A with repeat"""
	driver._repeatcount+=1
	dec = []
	if(driver._repeatcount < 10): return dec
	else: driver._repeatcount = 0
	for key in keys:
		if(key == 22): dec += ['dn']
		elif(key == 21): dec += ['up']
		elif(key == 24): dec += ['right']
		elif(key == 25): dec += ['left']
		elif(key == 23): dec += ['dn2']
		elif(key == 20): dec += ['up2']
		elif(key == 27): dec += ['left2']
		elif(key == 26): dec += ['right2']
	return dec

def brl_decode_key_names_A(keys,driver):
	"""translate key names for protocol A"""
	dec = []
	for key in keys:
		if(key == 22): dec += ['dn']
		elif(key == 21): dec += ['up']
		elif(key == 24): dec += ['right']
		elif(key == 25): dec += ['left']
		elif(key == 28): dec += ['l1']
		elif(key == 29): dec += ['l2']
		elif(key == 28 + driver._numcells * 2 + 4): dec += ['r1']
		elif(key == 29 + driver._numcells * 2 + 4): dec += ['r2']
		elif(key == 23): dec += ['dn2']
		elif(key == 20): dec += ['up2']
		elif(key == 27): dec += ['left2']
		elif(key == 26): dec += ['right2']
	return dec

def brl_decode_key_names_B(keys,driver):
	"""translate key names for protocol B"""
	dec = []
	for key in keys:
		if(key==31): dec += ['l1']
		elif(key == 30): dec += ['l2']
		elif(key == 39 + driver._numcells * 2): dec += ['r1']
		elif(key == 38 + driver._numcells * 2): dec += ['r2']
		elif(key == 18): dec += ['left']
		elif(key == 19): dec += ['right']
		elif(key == 22): dec += ['up']
		elif(key == 21): dec += ['dn']
		elif(key == 16): dec += ['left2']
		elif(key == 17): dec += ['right2']
		elif(key == 23): dec += ['up2']
		elif(key == 20): dec += ['dn2']
	return dec

def brl_decode_key_names_B_repeat(keys,driver):
	"""translate key names for protocol B with repeat"""
	driver._repeatcount += 1
	dec = []
	if(driver._repeatcount < 10): return dec
	else: driver._repeatcount = 0
	for key in keys:
		if(key == 18): dec += ['left']
		elif(key == 19): dec += ['right']
		elif(key == 22): dec += ['up']
		elif(key == 21): dec += ['dn']
		elif(key == 16): dec += ['left2']
		elif(key == 17): dec += ['right2']
		elif(key == 23): dec += ['up2']
		elif(key == 20): dec += ['dn2']
	return dec

        
def brl_join_keys(dec):
	"""join key names with comma, this is used for key combinations"""
	if(len(dec) == 1): return dec[0]
	elif(len(dec) == 3 and dec[0] == dec[1]): return dec[0] + "," + dec[2]
	elif(len(dec) == 3 and dec[0] == dec[2]): return dec[0] + "," + dec[1]
	elif(len(dec) == 2): return dec[1] + "," + dec[0]
	else: return ''


def brl_keyname_decoded(key,rest):
	"""convert index used by brxcom to keyname"""
	if(key == 11 or key == 9): return 'l1' + rest
	elif(key == 12 or key == 10): return 'l2' + rest
	elif(key == 13 or key == 15): return 'r1' + rest
	elif(key == 14 or key == 16): return 'r2' + rest
	
	elif(key == 3): return 'up' + rest
	elif(key == 7): return 'dn' + rest
	elif(key == 1): return 'left' + rest
	elif(key == 5): return 'right' + rest

	elif(key == 4): return 'up2' + rest
	elif(key == 8): return 'dn2' + rest
	elif(key == 2): return 'left2' + rest
	elif(key == 6): return 'rigth2' + rest
	else: return ''        
	
	
        


class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	"""papenmeier braille display driver.
	"""
	name = "papenmeier"
	description = _("Papenmeier BRAILLEX NEW")

	@classmethod
	def check(cls):
		"""should return false if there is a missing dependency"""
		return True


	def connectBrxCom(self):
		"""connect to brxcom server (the dlls are provided by papenmeier, but not yet released)"""
		try:
			brxcomkey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,"SOFTWARE\\FHP\\BrxCom")
			value, vtype = _winreg.QueryValueEx(brxcomkey, "InstallPath")
			_winreg.CloseKey(brxcomkey)
			self._brxnvda = c.cdll.LoadLibrary(value+"\\brxnvda.dll")
			if(self._brxnvda.brxnvda_init(str(value+"\\BrxCom.dll")) == 0):
				self._baud = 1 #prevent bluetooth from connecting
				self._numcells = self._brxnvda.brxnvda_numCells()
				self._voffset = self._brxnvda.brxnvda_numVertCells()
				log.info("Found Braille Display connected via BRXCom")
				self.startTimer()
				return None
		except:
			log.info("BRXCom is not installed")
			self._brxnvda = None

	def connectBluetooth(self):
		"""try to connect to bluetooth device first, bluetooth is only supported on Braillex Trio"""
		if(self._dev == None and self._baud == 0):
			for portInfo in sorted(hwPortUtils.listComPorts(onlyAvailable=True), key=lambda item: "bluetoothName" in item):
				port = portInfo["port"]
				hwID = portInfo["hardwareID"]
				if "bluetoothName" in portInfo:
					if(portInfo["bluetoothName"] == "braillex trio "):
						try:
							self._dev = serial.Serial(port, baudrate = 57600,timeout = BLUETOOTH_TIMEOUT, writeTimeout = BLUETOOTH_TIMEOUT)
							self._numcells = 40
							self._proto = 'B'
							log.info("connectBluetooth success")
						except:
							log.info("connectBluetooth failed")

	
	def connectUSB(self,devlist):
		"""try to connect to usb device,is triggered when BRXCOM is not installed and bluetooth
connection could not be established"""
		try:
			self._dev = ftdi2.open_ex(devlist[0])
			self._dev.set_baud_rate(self._baud)
			self._dev.inWaiting = self._dev.get_queue_status
			log.info("connectUSB success")
		except:
			log.info("connectUSB failed")
			
 
	def __init__(self): 
		"""initialize driver"""
		super(BrailleDisplayDriver, self).__init__()
		self._numcells = 0
		self._nlk = 0
		self._nrk = 0
		self._r1next = 'r1a'
		self.decodedkeys = []
		self._baud = 0
		self._dev = None
		self._proto = None
		self.connectBrxCom()
		if(self._baud == 1): return #brxcom is running, skip bluetooth and USB
		self._modstate = 0
		
			
		
		
		#try to connect to usb device,
		#if no usb device is found there may be a bluetooth device
		try:
			devlist = ftdi2.list_devices()
		except:
			devlist = []
		if(len(devlist)==0):
			self.connectBluetooth()
		else:
			self._baud = 57600
			self.connectUSB(devlist)
			if(self._dev == None):
				return None
			try:
				#request type of braille display
				self._dev.write(brl_auto_id())
				time.sleep(1)
				autoid=brl_poll(self._dev)
				if(autoid == ''):
					#no response, assume a Trio is connected
					self._baud = 115200
					self._dev.set_baud_rate(self._baud)
					self._dev.purge()        
					self._dev.read(self._dev.inWaiting())
					self._numcells = 40
					self._proto = 'B'
					self._voffset = 0
					#we don't use autoid here, because the trio might be switched off and other braille displays using the same baud rate do not exist
				else:
					if(len(autoid) != 8):
						return None
					autoid = struct.unpack('BBBBBBBB',autoid)
					if(autoid[3] == 0x35 and autoid[4] == 0x38):#EL80s
						self._numcells = 80
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL80s")
					elif(autoid[3]==0x35 and autoid[4]==0x3A):#EL70s
						self._numcells = 70
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL70s")
					elif(autoid[3]==0x35 and autoid[4]==0x35):#EL40s
						self._numcells = 40
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL40s")
					elif(autoid[3] == 0x35 and autoid[4] == 0x37):#EL66s
						self._numcells = 66
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL66s")
					elif(autoid[3] == 0x35 and autoid[4] == 0x3E):#EL20c
						self._numcells = 20
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL20c")
					elif(autoid[3] == 0x35 and autoid[4] == 0x3F):#EL40c
						self._numcells = 40
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL40c")
					elif(autoid[3] == 0x36 and autoid[4] == 0x30):#EL60c
						self._numcells = 60
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL60c")
					elif(autoid[3] == 0x36 and autoid[4] == 0x31):#EL80c
						self._numcells = 80
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 0
						log.info("Found EL80c")
					elif(autoid[3] == 0x35 and autoid[4] == 0x3b):#EL2D80s
						self._numcells = 80
						self._nlk = 1
						self._nrk = 1
						self._proto = 'A'
						self._voffset = 20
						log.info("Found EL2D80s")
					else:
						log.info('UNKNOWN BRAILLE')

			except:
				log.info('BROKEN PIPE - THIS SHOULD NEVER HAPPEN')
		if(self._numcells == 0): raise Exception('no device found')                
		#start keycheck timer
		self.startTimer()                                        
		
	def startTimer(self):
		"""start timers used by this driver"""
		self._keyCheckTimer = wx.PyTimer(self._handleKeyPresses)
		self._keyCheckTimer.Start(KEY_CHECK_INTERVAL)
		#the keycheck timer polls the braille display for keypresses
		self._bluetoothTimer = wx.PyTimer(self.connectBluetooth)
		self._bluetoothTimer.Start(BLUETOOTH_INTERVAL)
		#the bluetooth timer tries to reconnect if the bluetooth connection is lost 

	def stopTimer(self):
		"""stop all timers"""
		try:
			self._keyCheckTimer.Stop()
			self._bluetoothTimer.Stop()
		except:
			pass
		
		self._keyCheckTimer = None
		self._bluetoothTimer = None
		
		
		
		

	def terminate(self):
		"""free resources used by this driver"""
		try:
			fill = []
			for i in range(0,self._numcells): fill += [255]
			self.display(fill)
			self.stopTimer()
			if(self._dev!=None): self._dev.close()
			self._dev=None
			if(self._brxnvda): self._brxnvda.brxnvda_close()
			#free superclass resources
			super(BrailleDisplayDriver, self).terminate()
		except:
			self._dev=None        

	def _get_numCells(self):
		"""returns number of cells"""
		return self._numcells

	def display(self, cells):
		"""write to braille display"""
		if(self._brxnvda):
			newcells = ''
			for cell in cells: newcells += chr(cell)
			self._brxnvda.brxnvda_sendToDisplay(newcells)
		if(self._dev==None): return
		try:
			self._dev.write(brl_out(cells, self._nlk, self._nrk,self._voffset))
		except:
			self._dev.close()
			self._dev=None

	def executeGesture(self,gesture):
		"""executes a gesture"""
		if(gesture.id == 'r1'):  gesture.id = self.r1next()       
		if(len(gesture.id)): inputCore.manager.executeGesture(gesture)
		
	def _handleKeyPresses(self): 
		"""handles key presses and performs a gesture"""
		try:
			if(self._brxnvda):
				k = self._brxnvda.brxnvda_keyIndex()
				if(k!=-1):
					self.executeGesture(InputGesture(k,self))
				return
			if(self._dev==None and self._baud>0):
				try:
					devlist = ftdi2.list_devices()
					if(len(devlist)>0):
						self.connectUSB(devlist)
				except:
					return
			s = brl_poll(self._dev)
			if(len(s)):
				self._repeatcount=0
				ig = InputGesture(s,self)
				self.executeGesture(ig)                             
			else:
				if(len(self.decodedkeys)):
					ig = InputGesture(None,self)
					self.executeGesture(ig)
		except:
			if(self._dev!=None): self._dev.close()
			self._dev=None
	def r1next(self):
		"""select next key event for r1,used by the inputgesture class"""
		ret = self._r1next
		if(ret == 'r1a'): self._r1next = 'r1b'
		if(ret == 'r1b'): self._r1next = 'r1a'
		return ret

	#global gestures
	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(papenmeier):left",),
			"braille_scrollForward": ("br(papenmeier):right",),
			"braille_previousLine": ("br(papenmeier):up",),
			"braille_nextLine": ("br(papenmeier):dn",),
			"braille_routeTo": ("br(papenmeier):route",),

			"navigatorObject_moveToFlatReviewAtObjectPosition": ("br(papenmeier):r1a",),
			"braille_toggleTether": ("br(papenmeier):r2",),
			"review_currentCharacter": ("br(papenmeier):l1",),
			"reportFormatting": ("br(papenmeier):reportf",),
			"navigatorObject_toFocus":("br(papenmeier):r1b",),
			"navigatorObject_doDefaultAction": ("br(papenmeier):l2",),

			"navigatorObject_previous": ("br(papenmeier):left2",),
			"navigatorObject_next": ("br(papenmeier):right2",),
			"navigatorObject_parent": ("br(papenmeier):up2",),
			"navigatorObject_firstChild": ("br(papenmeier):dn2",),

			"title": ("br(papenmeier):l1,up",),
			"reportStatusLine": ("br(papenmeier):l2,dn",),
			
		}
	})

	

class InputGesture(braille.BrailleDisplayGesture):
	"""Input gesture for papenmeier displays"""

	source = BrailleDisplayDriver.name
	
	def __init__(self, keys,driver):
		"""create an input gesture and decode keys"""
		super(InputGesture, self).__init__()
		self.id=''

		if(keys == None):
			if(driver._proto == 'A'):
				self.id=brl_join_keys(brl_decode_key_names_A_repeat(driver.decodedkeys,driver))
			elif(driver._proto == 'B'):#trio
				self.id=brl_join_keys(brl_decode_key_names_B_repeat(driver.decodedkeys,driver))
			return

		if(driver._baud==1):#brxcom
			if(keys>255 and keys<512):
				self.routingIndex = keys-256-driver._voffset
				self.id = "route"
				return None
			elif(keys>511 and keys <786):
				braille.handler.routeTo(keys-512-driver._voffset)
				time.sleep(0.5)
				self.id='reportf'
				return None
			else:
				key1 = (keys & 0xFFFF0000) >> 16
				key2 = keys & 0x0000FFFF
				self.id=brl_keyname_decoded(key1, ',')+brl_keyname_decoded(key2, '')
				return None
		
		if(driver._proto == 'A'):#non trio
			decodedkeys = brl_decode_keys_A(keys[3:], 4, driver._voffset*2)
			
			
			
		elif(driver._proto=='B'):#trio
			decodedkeys = brl_decode_trio(keys)
			
		if(len(decodedkeys)==1 and decodedkeys[0]>=32 and decodedkeys[0]<32+driver._numcells*2):
			#routing keys
			self.routingIndex = (decodedkeys[0]-32)/2
			self.id = "route"
			if(decodedkeys[0] % 2 == 1):
				braille.handler.routeTo(self.routingIndex)
				time.sleep(0.5)
				self.id='reportf'
		#other keys                
		elif(len(decodedkeys) > 0 and len(decodedkeys) >= len(driver.decodedkeys)):
			driver.decodedkeys += decodedkeys
		elif(len(decodedkeys) == 0 and len(driver.decodedkeys)>0):
			if(driver._proto == 'A'):
				self.id=brl_join_keys(brl_decode_key_names_A(driver.decodedkeys,driver))
			elif(driver._proto == 'B'):#trio
				self.id=brl_join_keys(brl_decode_key_names_B(driver.decodedkeys,driver))                             
			driver.decodedkeys=[]

			      
			   
		       
		       
	      
